package ch.zhaw.pm2.multichat.server;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Map;
import java.util.Objects;
import java.util.Scanner;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;



public class ServerConnectionHandler extends ConnectionHandler implements Runnable{
    private static final AtomicInteger connectionCounter = new AtomicInteger(0);
    private final int connectionId = connectionCounter.incrementAndGet();
    private final Map<String,ServerConnectionHandler> connectionRegistry;

    private ReentrantLock mutex;

    private Condition nameComplete;

    private String userName = "Anonymous-"+connectionId;
    private State state = NEW;

    @Override
    public void run() {
        startReceiving();
    }

    public ServerConnectionHandler(NetworkHandler.NetworkConnection<String> connection,
                                   Map<String,ServerConnectionHandler> registry, ReentrantLock mutex, Condition nameComplete) {
        super();
        setConnection(connection);
        Objects.requireNonNull(connection, "Connection must not be null");
        Objects.requireNonNull(registry, "Registry must not be null");
        this.connectionRegistry = registry;
        this.mutex = mutex;
        this.nameComplete = nameComplete;
    }

    public String getUserName() {
        return this.userName;
    }

    public State getState() {
        return state;
    }

    private void startReceiving() {
            System.out.println("Starting Connection Handler for new User");
            try {
                System.out.println("Start receiving data...");
                while (getConnection().isAvailable() && !(state == ERROR)) {
                    String data = getConnection().receive();
                    processData(data);
                }
                System.out.println("Stopped recieving data");
            } catch (SocketException e) {
                System.out.println("Connection terminated locally");
                connectionRegistry.remove(userName);
                System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
            } catch (EOFException e) {
                System.out.println("Connection terminated by remote");
                connectionRegistry.remove(userName);
                System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
            } catch (IOException e) {
                System.err.println("Communication error: " + e);
            } catch (ClassNotFoundException e) {
                System.err.println("Received object of unknown type: " + e.getMessage());
            }
        if (state == ERROR) {
            System.out.println("Stopping Connection Handler for Rejected Client");
        } else {
            System.out.println("Stopping Connection Handler for " + userName);
        }

    }

    private void stopReceiving() {
        System.out.println("Closing Connection Handler for " + userName);
        try {
            System.out.println("Stop receiving data...");
            getConnection().close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e.getMessage());
        }
        System.out.println("Closed Connection Handler for " + userName);
    }

    private void processData(String data)  {
        try {
            Scanner scanner = new Scanner(data);
            StringBuilder sender = new StringBuilder();
            StringBuilder reciever = new StringBuilder();
            StringBuilder type = new StringBuilder();
            StringBuilder payload = new StringBuilder();
            super.processData(scanner,sender,reciever,type,payload);

            // dispatch operation based on type parameter
            if (type.toString().equals(getDataTypeConnect())) {
                if (this.state != NEW) throw new ChatProtocolException("Illegal state for connect request: " + state);
                if (sender.toString().isBlank()) sender.append(this.userName);
                if (connectionRegistry.containsKey(sender.toString())) {
                    mutex.lock();
                    try {
                        state = ERROR;
                        nameComplete.signal();
                    }
                    finally {
                        mutex.unlock();
                    }
                    throw new ChatProtocolException("User name already taken: " + sender);
                }
                mutex.lock();
                try {
                    this.userName = sender.toString();
                    nameComplete.signal();
                }
                finally {
                    mutex.unlock();
                }
                connectionRegistry.put(userName, this);
                sendData(USER_NONE, userName, getDataTypeConfirm(), "Registration successfull for " + userName);
                this.state = CONNECTED;
            } else if (type.toString().equals(getDataTypeConfirm())) {
                System.out.println("Not expecting to receive a CONFIRM request from client");
            } else if (type.toString().equals(getDataTypeDisconnect())) {
                if (state == DISCONNECTED)
                    throw new ChatProtocolException("Illegal state for disconnect request: " + state);
                if (state == CONNECTED) {
                    connectionRegistry.remove(this.userName);
                }
                sendData(USER_NONE, userName, getDataTypeConfirm(), "Confirm disconnect of " + userName);
                this.state = DISCONNECTED;
                this.stopReceiving();
            } else if (type.toString().equals(getDataTypeMessage())) {
                if (state != CONNECTED) throw new ChatProtocolException("Illegal state for message request: " + state);
                if (USER_ALL.equals(reciever.toString())) {
                    for (ServerConnectionHandler handler : connectionRegistry.values()) {
                        handler.sendData(sender.toString(), reciever.toString(), type.toString(), payload.toString());
                    }
                } else {
                    ServerConnectionHandler handler = connectionRegistry.get(reciever.toString());
                    if (handler != null) {
                        handler.sendData(sender.toString(), reciever.toString(), type.toString(), payload.toString());
                        if(!reciever.toString().equals(sender.toString())){
                            sendData(sender.toString(), reciever.toString(), type.toString(), payload.toString()); //send message to sender if it's a direct message and sender is not receiver.
                        }
                    } else {
                        this.sendData(USER_NONE, userName, getDataTypeError(), "Unknown User: " + reciever);
                    }
                }
            } else if (type.toString().equals(getDataTypeError())) {
                System.err.println("Received error from client (" + sender + "): " + payload);
            } else {
                System.err.println("Unknown data type received: " + type);

            }
        } catch(ChatProtocolException e) {
            System.out.println("Error while processing data " + e.getMessage());
            sendData(USER_NONE, userName, getDataTypeError(), e.getMessage());
        }
    }
}
